/*
 * Copyright (c) 2010 Samsung Electronics Co., Ltd.
 *              http://www.samsung.com/
 *
 * S5PC110 - LCD Driver for U-Boot
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License version 2 as
 * published by the Free Software Foundation.
 */

#include <common.h>
#include <asm/io.h>
#include <asm/types.h>
#include <lcd.h>
#include <video_font.h>
#include <lcd.h>
#include "FONT.h"

DECLARE_GLOBAL_DATA_PTR;

#define LCD_BGCOLOR		0x1428A0
#define LCD_BPP                           LCD_COLOR8


static unsigned int gFgColor = 0xFF;
static unsigned int gLeftColor = LCD_BGCOLOR;

void (*PutPixel)(u32 x, u32 y, u32 c);

#define Inp32(_addr)		readl(_addr)
#define Outp32(addr, data)	(*(volatile u32 *)(addr) = (data))
#define Delay(_a)		udelay(_a*1000)


#if defined(CFG_LCD_LB070WV6)
/*
 * LB070WV6
 */

#define LCD_WIDTH		800
#define LCD_HEIGHT		480
#define LCD_VBPD		23
#define LCD_VFPD		22
#define LCD_VSPW		1
#define LCD_HBPD		46
#define LCD_HFPD		210
#define LCD_HSPW		3

#define LCD_ORG_WIDTH   96  /* 6 column */
#define LCD_ORG_HEIGHT  16

u32 g_fbuffer_data_LCD[LCD_WIDTH * LCD_HEIGHT];

vidinfo_t panel_info = {
        .vl_col = 800,
        .vl_row = 480,
        .vl_bpix = 3,
};

int lcd_line_length;
int lcd_color_fg;
int lcd_color_bg;

void *lcd_base;                         /* Start of framebuffer memory  */
void *lcd_console_address;              /* Start of console buffer      */

short console_col = 5;
short console_row = 5;

void LCD_Initialize_LB070WV6(void)
{
	u32 uFbAddr = CFG_LCD_FBUFFER;

	u32 i,uTmpReg;
	u32* pBuffer = (u32*)uFbAddr;

	printf("%s %d \n",__func__,__LINE__);

	LCD_setleftcolor(0x0);
	LCD_setfgcolor(0x0);
	LCD_setprogress(0);


	uTmpReg = Inp32(0xE0107008);		// need to be changed later (09.01.23 WJ.Kim)	
	uTmpReg = (uTmpReg & ~(0x3)) | 0x2 ;
	Outp32(0xE0107008, uTmpReg);

	uTmpReg = Inp32(0xf800027c);
	uTmpReg = (uTmpReg & ~(0x3)) | 0 ;
	Outp32(0xf800027c,uTmpReg);


	Outp32(0xE0200120, 0x22222222);	//set GPF0 as LVD_HSYNC,VSYNC,VCLK,VDEN,VD[3:0]
	Outp32(0xE0200128,0x0);			//set pull-up,down disable
	Outp32(0xE0200140, 0x22222222);	//set GPF1 as VD[11:4]
	Outp32(0xE0200148,0x0);			//set pull-up,down disable
	Outp32(0xE0200160, 0x22222222);	//set GPF2 as VD[19:12]
	Outp32(0xE0200168,0x0);			//set pull-up,down disable
	Outp32(0xE0200180, 0x00002222);	//set GPF3 as VD[23:20]
	Outp32(0xE0200188,0x0);			//set pull-up,down disable

	Outp32(0xe02000A0, 0x11111111);	//set GPD0 as Backlight
	Outp32(0xe02000A8,0x0);			//set pull-up,down disable
	Outp32(0xe02000A4,(1<<0));

	Outp32(0xE0200080, 0x11111111);	//set GPC1 as LCD_PWREN
	Outp32(0xE0200088,0x0);			//set pull-up,down disable
	Outp32(0xE0200084,(1<<4));

	Outp32(0xf8000004, 0x60);
	Outp32(0xf8000010, ((LCD_VBPD-1)<<16)|((LCD_VFPD-1)<<8)|(LCD_VSPW-1));
	Outp32(0xf8000014, ((LCD_HBPD-1)<<16)|((LCD_HFPD-1)<<8)|(LCD_HSPW-1));
	Outp32(0xf8000170, 0x0);
	Outp32(0xf8000018, ((LCD_HEIGHT-1)<<11)|(LCD_WIDTH-1));
	Outp32(0xf8000000, 0x0);
	Outp32(0xf8000000, 0x190);
	//Outp32(0xf8000000, 0x110);
	Outp32(0xf8000130, 0x20); //Window 0 - Enable
#if 0
	Outp32(0xf8000020, 0x0);
	Outp32(0xf8000024, 0x0);
	Outp32(0xf8000028, 0x0);
	Outp32(0xf800002c, 0x0);
	Outp32(0xf8000030, 0x0);
	Outp32(0xf8000034, 0x0);
	Outp32(0xf8000180, 0x0);
	Outp32(0xf8000184, 0x0);
	Outp32(0xf8000188, 0x0);
	Outp32(0xf800018c, 0x0);
	Outp32(0xf8000190, 0x0);
	Outp32(0xf8000140, 0x0);
	Outp32(0xf8000148, 0x0);
	Outp32(0xf8000150, 0x0);
	Outp32(0xf8000158, 0x0);
	Outp32(0xf8000058, 0x0);
	Outp32(0xf8000208, 0x0);
	Outp32(0xf800020c, 0x0);
	Outp32(0xf8000068, 0x0);
	Outp32(0xf8000210, 0x0);
	Outp32(0xf8000214, 0x0);
	Outp32(0xf8000078, 0x0);
	Outp32(0xf8000218, 0x0);
	Outp32(0xf800021c, 0x0);
	Outp32(0xf8000088, 0x0);
	Outp32(0xf8000220, 0x0);
	Outp32(0xf8000224, 0x0);
#endif
	Outp32(0xf8000260, 0x1);
	Outp32(0xf8000034, 0x0);
	//FrameBuffer Start/End Address Setting
	Outp32(0xf80000a4, uFbAddr + 0x0);
	Outp32(0xf80000d4, uFbAddr + (LCD_WIDTH*LCD_HEIGHT*4));
	Outp32(0xf80000a0, uFbAddr + 0x0);
	Outp32(0xf80000d0, uFbAddr + (LCD_WIDTH*LCD_HEIGHT*4));
	Outp32(0xf80020a0, uFbAddr + 0x0);
	Outp32(0xf80020d0, uFbAddr + (LCD_WIDTH*LCD_HEIGHT*4));
	Outp32(0xf8000100, (LCD_WIDTH*4));
	Outp32(0xf8000020, 0x802c);
	Outp32(0xf8000040, 0x0);
	Outp32(0xf8000044, ((LCD_WIDTH-1)<<11)|(LCD_HEIGHT-1));
	Outp32(0xf8000200, 0xffffff);
	Outp32(0xf8000204, 0xffffff);
	Outp32(0xf8000020, 0x802d);
	Outp32(0xf8000034, 0x1);

	Outp32(0xf8000000, 0x193);
	//Outp32(0xf8000000, 0x113);
}
#else
// "No LCD Type is defined!"
#endif

void _PutPixel32Bit(u32 x, u32 y, u32 c)
{
	u32* pBuffer = (u32*)CFG_LCD_FBUFFER;
    if(x < LCD_WIDTH && y < LCD_HEIGHT)
        //*(u32 *)(g_fbuffer_data_LCD + y * LCD_ORG_WIDTH + x) = c;
        *(u32 *)(pBuffer + y * LCD_WIDTH + x) = c;
}
/*
 */

#define N_BLK_VERT      2
#define N_BLK_HOR       3

static int test_colors[N_BLK_HOR*N_BLK_VERT] = {
        CONSOLE_COLOR_RED,      CONSOLE_COLOR_GREEN,    CONSOLE_COLOR_YELLOW,
        CONSOLE_COLOR_BLUE,     CONSOLE_COLOR_MAGENTA,  CONSOLE_COLOR_CYAN,
};

static void test_pattern (void)
{
        ushort v_max  = panel_info.vl_row;
        ushort h_max  = panel_info.vl_col;
        ushort v_step = (v_max + N_BLK_VERT - 1) / N_BLK_VERT;
        ushort h_step = (h_max + N_BLK_HOR  - 1) / N_BLK_HOR;
        ushort v, h;
        //uchar *pix = (uchar *)lcd_base;

        printf ("[LCD] Test Pattern: %d x %d [%d x %d]\n",
                h_max, v_max, h_step, v_step);

        /* WARNING: Code silently assumes 8bit/pixel */
        for (v=0; v<v_max; ++v) {
                uchar iy = v / v_step;
                for (h=0; h<h_max; ++h) {
                        uchar ix = N_BLK_HOR * iy + (h/h_step);
                        //*pix++ = test_colors[ix];
						PutPixel(v,h,test_colors[ix]);
                }
        }
}

void lcd_ctrl_init (void *lcdbase)
{
	lcd_line_length = (panel_info.vl_col * NBITS (panel_info.vl_bpix)) / 8;
	lcd_base = (void *)(gd->fb_base);
	LCD_Initialize_LB070WV6();
}

void LCD_testpattern(void)
{
	test_pattern();
}
void LCD_turnon(void)
{
	lcd_line_length = (panel_info.vl_col * NBITS (panel_info.vl_bpix)) / 8;
	//printf("lcd_line_length = %d \n",lcd_line_length);
    PutPixel=_PutPixel32Bit;
 	lcd_base = (u32*)CFG_LCD_FBUFFER;
	LCD_Initialize_LB070WV6();
}

void LCD_setfgcolor(unsigned int color)
{
	lcd_color_fg = color;
}

void LCD_setleftcolor(unsigned int color)
{
	lcd_color_bg = color;
}

void LCD_setprogress(int percentage)
{
	u32 i, j;
	u32* pBuffer = (u32*)CFG_LCD_FBUFFER;
#if 1
	for (i=0; i < (LCD_HEIGHT/100)*percentage; i++)
	{
		for (j=0; j < LCD_WIDTH; j++)
		{
			*pBuffer++ = lcd_color_fg;
		}
	}

	for (; i < LCD_HEIGHT; i++)
	{
		for (j=0; j < (LCD_WIDTH >> 5); j++)
		{
			*pBuffer++ = lcd_color_bg;
		}
		for (; j < LCD_WIDTH; j++)
		{
			*pBuffer++ = lcd_color_bg;
		}
	}
#endif

	//for(i=0;i<LCD_HEIGHT;i++)
	//	for(j=0;j<LCD_WIDTH;j++)
	//		*pBuffer++ = gFgColor;
}


void lcd_setcolreg (ushort regno, ushort red, ushort green, ushort blue)
{
	printf("%s..\n",__func__);
 
}
void lcd_enable(void)
{
	printf("%s..\n",__func__);

}

void lcd_disable(void)
{
	printf("%s..\n",__func__);

}

void lputs(u32 x, u32 y, u32 c, char *str)
{
	int i,j,k;
	int oneLine;

	for(i=0;i<strlen(str);i++)
	{
		for(j=0; j<16; j++)
		{
			if ((str[i]>=0x20)&&(str[i]<0x80))
				oneLine=chr_font8X16[str[i]-0x20][j];
			else
				oneLine=0;

			for(k=0;k<8;k++)
			{
				if(oneLine&(1<<k))
					PutPixel(x+7-k,y+j,c);
			}
		}
		x+=8;
	}
}
void lprintf(u32 x, u32 y, u32 c, const char *fmt,...)
{
	va_list ap;
	char string[256];

	va_start(ap,fmt);
	vsprintf(string,fmt,ap);
	lputs(x,y,c,string);
	va_end(ap);
}

void LCD_set_fb_tmp_test(ulong command, ulong value1, ulong value2, ulong value3)
{
    int i;

	printf ("LCD_set_fb_tmp_test Command:%d, value1:%d, value2:%d, value3:%d\n",
            command, value1, value2, value3);

    switch(command)
    {
    case 0:
        printf ("Command 0 - Test Value\n");
		LCD_turnon();
		LCD_setfgcolor(value1);
		LCD_setleftcolor(value2); //CRZ icanjji
		for(i = 0 ; i < 100; i++)
			LCD_setprogress(i);
        break;

    case 1:
    	printf ("Command q - char print\n");
        lprintf(0, 0, value1, "1234567890AB");
        break;

    case 2:
    	printf ("Command 2 - test_pattern \n");
		test_pattern();
        break;

    case 3:
    	printf ("Command 3 - test_pattern \n");
#if 1
		LCD_turnon();
		LCD_setfgcolor(value1);
		LCD_setleftcolor(value2); //CRZ icanjji
		for(i = 0 ; i < 100; i++)
			LCD_setprogress(i);
        lprintf(420, 400, value3, "1234567890AB");
#else 
		LCD_turnon();
		LCD_setfgcolor(0xff0000);
		LCD_setleftcolor(0x0); //CRZ icanjji
		LCD_setprogress(100);
		lprintf(400, 400, 0xffffff, "update u-boot.bin....ing");
		
		LCD_setprogress(100);
		lprintf(400, 400, 0xffffff, "update zImage,ramdisk-uboot.img....ing");
#endif
        break;
    default:
    	printf ("lcd_control Command fail\n");
        break;
    }

}

int
do_lcd_control (cmd_tbl_t *cmdtp, int flag, int argc, char *argv[])
{
	ulong command, value1, value2, value3;
	char **av;

//    lcd_width_height_setting();

	av = argv + 1;

	command = simple_strtoul(*av++, NULL, 16);
	value1 = simple_strtoul(*av++, NULL, 16);
	value2 = simple_strtoul(*av++, NULL, 16);
	value3 = simple_strtoul(*av++, NULL, 16);

    //LCD_set_fb_initialData();
    LCD_turnon();
    LCD_set_fb_tmp_test(command, value1, value2, value3);
	return 0;
}

U_BOOT_CMD(
	lcd,	5,		1,	do_lcd_control,
	"lcd [cmd] [color1-fg] [color2-bg] [color3-font] - lcd control\n",
	NULL
);
